/****
file:       registry.cpp

  purpose:    module initialization and cleanup,
  Main Message Handler
  Bang Commands

Syntax is :

(* / !)registryread LSvariable Root Key Value (use quotes or [] to encapsulate each field)

v 0.9
=====
*sigh* Pointers - I forgot the basics :)
Spaced values failed to work properly. Fixed.
Memory leak identified by ilmcuts. Fixed.

v 0.6
=====
ilmcuts pointed out a silly flaw. Fixed :)
Rewrote the registry code using Tobbe's code as a base to work from.
I really must knock some C# wrappers up to make my life easier :)

v 0.5
=====
Far, far too much time spent to get this working. *sigh*

- Resolved the linker problems with VS2005EE. Have to manually add advapi32.lib, user32.lib and lsapi.lib.
 - ATL also missing from the Express Edition, helpfully.
- Fixed InitModuleEx() problem - forgot to include registry.h *smack around head*
- Credit to Richard Reynolds for much patient assistance :
  - Found out that Windows functions require Unicode, so introduced conversion code to make this happen (ATL).
  - He also spotted numerous silly errors with the conversion code :)
- jugg pointed out that GetToken requires a char array to drop (a) token(s) into.

****/

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Includes
//
// #include <stdafx.h>
#include <lsapi/lsapi.h>
#include <windows.h>
#include <string>
#include "registry.h"

// global variables
extern BOOL userParanoia = FALSE;

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ErrorMsg
//
// Error handler
//
void ErrorMsg(LPCSTR pszMessage)
{
    MessageBox(NULL, pszMessage, "registryread error", MB_ICONEXCLAMATION | MB_TOPMOST | MB_SETFOREGROUND);
}


void ErrorMsgf(LPCSTR pszFormat, ...)
{
    char szMessage[MAX_LINE_LENGTH];
    va_list argList;
    va_start(argList, pszFormat);
    vsprintf(szMessage, pszFormat, argList);
    va_end(argList);
    // check to see if logging is enabled to determine how to report error.
    if(GetRCString("LSLogFile",NULL,"",MAX_LINE_LENGTH)) {
        LSLog(1, "registryread", szMessage); // LSLog is undocumented - guessing at the moment for log level.
    } else {
        ErrorMsg(szMessage);
    }
}

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// ProcessRegistryReadString
//
// Parses input line and attempts to export variable
//
// expecting *registry evar HKxx key value or *registry [evar] [HKxx] [key] [value]

bool ProcessRegistryReadString(LPCSTR pszLine)
{
	if (userParanoia == TRUE) // stop before we start when user is paranoid :)
	{
		ErrorMsg("Not reading registry due to user request.");
		return 1; // wondering if this should be 1 or 0 since it is not really an error in the classic sense.
	}

	// OK let's start by pulling in the information we need from pszLine using GetToken.
	// Once we have a stable syntax, this should have a counter added to allow errors to
	// be gracefully handled when too few/many arguments have been provided.

	// Name of LS variable to assign value to if all registry stuff works.
	CHAR szmyVar[MAX_LINE_LENGTH] = { 0 }; // destination variable
	if (!GetToken(pszLine, szmyVar, &pszLine, TRUE))
	{
		ErrorMsg("Read failed at token 1. Aborting.");
		return 1;
	}

	HKEY szBaseReg;
	CHAR szTemp[MAX_LINE_LENGTH] = { 0 };	// Registry root (HKEY_LOCAL_MACHINE, etc.)
	if (!GetToken(pszLine, szTemp, &pszLine, TRUE))
	{
		ErrorMsg("Read failed at token 2. Aborting.");
		return 1;
	}
	else
	{
		if (stricmp(szTemp, "HKCU") == 0)
			szBaseReg = HKEY_CURRENT_USER;
		else if (stricmp(szTemp, "HKLM") == 0)
			szBaseReg = HKEY_LOCAL_MACHINE;
		else
			return 1;
	}

	// Registry key leading to value.
	CHAR szKeyReg[MAX_LINE_LENGTH] = { 0 };
	if (!GetToken(pszLine, szKeyReg, &pszLine, TRUE))
	{
		ErrorMsg("Read failed at token 3. Aborting.");
		return 1;
	}

	// Registry value we are interested in
	CHAR szValueReg[MAX_LINE_LENGTH] = { 0 };
	if (!GetToken(pszLine, szValueReg, NULL , TRUE))
	{
		ErrorMsg("Read failed at token 4. Aborting.");
		return 1;
	}

	if (!keyExist(szBaseReg, szKeyReg))
		return 1;

	string* szReadValue = queryAKey(szBaseReg, szKeyReg, szValueReg);

	string szReadValueChar = *szReadValue;
	szReadValueChar.insert(0,"\"");
	szReadValueChar.append("\"");
	const char* szValue = szReadValueChar.c_str();
	LSSetVariable(szmyVar, szValue);

//	LSSetVariable(szmyVar, (*szReadValue).c_str());

	delete szReadValue;
	return 0;
}

// Pinched from Tobbe's WinReg.cpp/.h package - thanks! Massaged slightly to fit here.
/**
 * keyExist
 *
 * This function will check if a key exist
 *
 * @parameters
 * iRootKey - HKEY_CLASSES_ROOT, HKEY_CURRENT_USER,
 *			  HKEY_lOCAL_MACHINE, HKEY_USERS, etc
 * sRegKeyPath - The path to the key you want to check
 *
 * @return value
 * Boolean, true if the key exist, false if it doesn't
 *
 * @example
 * keyExist(HKEY_CURRENT_USER, "Software\Company\Stuff")
 */
bool keyExist(HKEY hRootKey, const string &sRegKeyPath)
{
	HKEY hKey = NULL;

	//Query the key path
	lRetVal = RegOpenKey(hRootKey, sRegKeyPath.c_str(), &hKey);

	//Always close the handle in the registry. We do not want to corrupt these files.
	lRetVal = RegCloseKey(hKey);

	return hKey != NULL;
}

/**
 * queryAKey
 *
 * This function will query (read) a sub key value
 *
 * @parameters
 * lngRootKey - HKEY_CLASSES_ROOT, HKEY_CURRENT_USER,
 *              HKEY_lOCAL_MACHINE, HKEY_USERS, etc
 * strRegKeyPath - The path to the key you want to query
 * strRegSubKey - The name of the subkey you want to query
 *
 * @return value
 * Variant, we don't know what we'll be retrieving, but it'll be the value of the SubKey
 *
 * @example
 * regQueryAKey(HKEY_CURRENT_USER, "Software\Company\Stuff", "DataKey")
 */
string* queryAKey(HKEY hRootKey, const string &sRegKeyPath, const string &sRegSubKey)
{
	HKEY hKey = NULL;
	DWORD dwDataType;
	DWORD dwSize;

	//Query the key path
	lRetVal = RegOpenKey(hRootKey, sRegKeyPath.c_str(), &hKey);

	//If no key handle was found then there is no key. Leave here.
	if (hKey == NULL)
	{
		lRetVal = RegCloseKey(hKey);   // always close the handle
		return new string("");
	}

	//Query the registry and determine the data type and the size of the data
	lRetVal = RegQueryValueEx(hKey, sRegSubKey.c_str(), NULL, &dwDataType, NULL, &dwSize);

	//If no key handle was found then there is no key. Leave.
	if (hKey == NULL)
	{
		lRetVal = RegCloseKey(hKey); // always close the handle
		return new string("");
	}

	//Allocate space for the data
	BYTE *bData = new BYTE[dwSize + 1];

	lRetVal = RegQueryValueEx(hKey, sRegSubKey.c_str(), NULL, NULL, bData, &dwSize);

	//If NOT a successful call then leave
	if (lRetVal != ERROR_SUCCESS)
	{
		lRetVal = RegCloseKey(hKey);
		return new string("");
	}

	string *sReturnVal = new string();

	//Make the API call to query the registry based on the type of data.
	switch (dwDataType)
	{
		case REG_SZ: //String data (most common)
		case REG_EXPAND_SZ:
		{
			//Convert the data to a string
			ostringstream oss;
			oss << bData;
			string sData = oss.str();

			delete [] bData;

			// Strip out the string data
			sData = sData.substr(0, sData.find_first_of('\0'));

			*sReturnVal = sData;

			break;
		}
		case REG_DWORD: //Numeric data (Integer)
		{
			DWORD dw;
			memcpy(&dw, bData, sizeof(DWORD));

			delete [] bData;

			//Save the captured data
			ostringstream oss;
			oss << dw;
			string sData = oss.str();

			// Strip out the string data
			sData = sData.substr(0, sData.find_first_of('\0'));

			*sReturnVal = sData;

			break;
		}
		default: //unknown
		{
			*sReturnVal = "";
			break;
		}
	}

	//Always close the handle in the registry. We do not want to corrupt these files.
	lRetVal = RegCloseKey(hKey);

	return sReturnVal;
}

// end of Tobbe's stuff :)

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// BangRegistryRead
//
// !bang analog to *registryread
//
void BangRegistryRead( HWND /*hCaller*/, LPCSTR szArgs)
{
	ProcessRegistryReadString(szArgs);
	// ErrorMsg(szArgs);
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// initModuleEx
//
// Module Initialization
//
int initModuleEx(IN HWND /* hLiteStep */, IN HINSTANCE /* hInstance */,
                 IN LPCSTR /* szLSpath */)
{
    AddBangCommand("!registryread", BangRegistryRead);
    
    FILE* fpStep = LCOpen(NULL);
    
    if (fpStep)
    {
        char szLineBuffer[MAX_LINE_LENGTH] = { 0 };
        
        while (LCReadNextConfig(fpStep, "*registryread", szLineBuffer, MAX_LINE_LENGTH))
        {
            LPCSTR pszNextToken = NULL;
            
            // Remove the *registryread token
            if (GetToken(szLineBuffer, NULL, &pszNextToken, TRUE))
            {
                ProcessRegistryReadString(pszNextToken);
            }
            else
            {
                ErrorMsgf("Can't parse line: %s", szLineBuffer);
            }
        }

       if(GetRCString("registryparanoia",NULL,"",MAX_LINE_LENGTH))
	   userParanoia = TRUE;
        
        LCClose(fpStep);

    }
    return 0;
}

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// quitModule
//
// module cleanup
//
void quitModule(IN HINSTANCE /* hInstance */)
{
    RemoveBangCommand("!registryread");
}


//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// DllMain
//
// DLL entry point
//
BOOL APIENTRY DllMain(IN HINSTANCE hInstance, IN DWORD dwReason,
                      LPVOID /* pvReserved */)
{
    if (dwReason == DLL_PROCESS_ATTACH)
    {
        DisableThreadLibraryCalls(hInstance);
    }
    
    return TRUE;
}